from commodutil import stats
from commodutil import forwards
import unittest
import os
import pandas as pd


class TestForwards(unittest.TestCase):

    def test_curve_zscore(self):
        dirname, filename = os.path.split(os.path.abspath(__file__))
        cl = pd.read_csv(os.path.join(dirname, 'test_cl.csv'), index_col=0, parse_dates=True, dayfirst=True)
        contracts = cl.rename(columns={x: pd.to_datetime(forwards.convert_contract_to_date(x)) for x in cl.columns})
        hist = contracts[['2020-01-01']].dropna()

        fwd = contracts[['2020-01-01']]

        res = stats.curve_seasonal_zscore(hist, fwd)

        self.assertAlmostEqual(res['zscore']['2019-01-02'], 0.92, 2)

    def test_reindex_zscore(self):
        dirname, filename = os.path.split(os.path.abspath(__file__))
        cl = pd.read_csv(os.path.join(dirname, 'test_cl.csv'), index_col=0, parse_dates=True, dayfirst=True)
        contracts = cl.rename(columns={x: pd.to_datetime(forwards.convert_contract_to_date(x)) for x in cl.columns})

        q = forwards.quarterly_contracts(contracts)
        q = q[[x for x in q.columns if 'Q1' in x]]

        res = stats.reindex_zscore(q)
        self.assertIsNotNone(res)


if __name__ == '__main__':
    unittest.main()


import re
import datetime 

curmon = datetime.datetime.now().month
curyear = datetime.datetime.now().year
curmonyear = datetime.datetime(curyear, curmon, 1)
curmonyear_str = '%s-%s' % (curyear, curmon) # get pandas time filtering

nextyear = curyear + 1


def find_year(df, use_delta=False):
    """
    Given a dataframe find the years in the column headings. Return a dict of colname to year
    eg { 'Q1 2016' : 2016, 'Q1 2017' : 2017
    """
    res = {}
    for colname in df:
        colregex = re.findall('\d\d\d\d', str(colname))
        colyear = None
        if len(colregex) >= 1:
            colyear = int(colregex[0])

        if colyear:
            res[colname] = colyear
            if colyear and use_delta:
                delta = colyear - curyear
                res[colname] = delta
        else:
            res[colname] = colname

    return res



"""
Utility for forward contracts
"""
import re
import pandas as pd
from calendar import month_abbr
from commodutil import dates

futures_month_conv = {
    1: "F",
    2: "G",
    3: "H",
    4: "J",
    5: "K",
    6: "M",
    7: "N",
    8: "Q",
    9: "U",
    10: "V",
    11: "X",
    12: "Z"
}

futures_month_conv_inv = {v: k for k, v in futures_month_conv.items()}


def convert_contract_to_date(contract):
    """
    Given a string like FB_2020J return 2020-01-01
    :param contract:
    :return:
    """
    c = re.findall('\d\d\d\d\w', contract)
    if len(c) > 0:
        c = c[0]
    d = '%s-%s-1' % (c[:4], futures_month_conv_inv.get(c[-1], 0))
    return d


def time_spreads_monthly(contracts, m1, m2):
    """
    Given a dataframe of daily values for monthly contracts (eg Brent Jan 15, Brent Feb 15, Brent Mar 15)
    with columns headings as '2020-01-01', '2020-02-01'
    Return a dataframe of time spreads  (eg m1 = 12, m2 = 12 gives Dec-Dec spread)
    """

    cf = [x for x in contracts if x.month == m1]
    dfs = []

    for c1 in cf:
        year1, year2 = c1.year, c1.year
        if m1 == m2:
            year2 = year1 + 1
        c2 = [x for x in contracts if x.month == m2 and x.year == year2]
        if len(c2) == 1:
            c2 = c2[0]
            s = contracts[c1] - contracts[c2]
            s.name = year1
            dfs.append(s)

    res = pd.concat(dfs, 1)
    res = res.dropna(how='all', axis='rows')
    return res


def time_spreads_quarterly(contracts, m1, m2):
    """
    Given a dataframe of daily values for monthly contracts (eg Brent Jan 15, Brent Feb 15, Brent Mar 15)
    with columns headings as '2020-01-01', '2020-02-01'
    Return a dataframe of time spreads  (eg m1 = Q1, m2 = Q2 gives Q1-Q2 spread)
    """

    qtrcontracts = quarterly_contracts(contracts)
    qtrcontracts_years = dates.find_year(qtrcontracts)
    cf = [x for x in qtrcontracts if x.startswith(m1)]
    dfs = []

    for c1 in cf:
        year1, year2 = qtrcontracts_years[c1], qtrcontracts_years[c1]
        if int(m1[-1]) >= int(m2[-1]):  # eg Q1-Q1 or Q4-Q1, then do Q419 - Q120 (year ahead)
            year2 = year1 + 1
        c2 = [x for x in qtrcontracts if x.startswith(m2) and qtrcontracts_years[x] == year2]
        if len(c2) == 1:
            c2 = c2[0]
            s = qtrcontracts[c1] - qtrcontracts[c2]
            s.name = year1
            dfs.append(s)

    res = pd.concat(dfs, 1)
    res = res.dropna(how='all', axis='rows')
    return res


def fly(contracts, m1, m2, m3):
    """
    Given a dataframe of daily values for monthly contracts (eg Brent Jan 15, Brent Feb 15, Brent Mar 15)
    with columns headings as '2020-01-01', '2020-02-01'
    Return a dataframe of flys  (eg m1 = 1, m2 = 2, m3 = 3 gives Jan/Feb/Mar fly)
    """

    cf = [x for x in contracts if x.month == m1]
    dfs = []
    for c1 in cf:
        year1, year2, year3 = c1.year, c1.year, c1.year
        # year rollover
        if m2 < m1:  # eg dec/jan/feb, make jan y+1
            year2 = year2 + 1
        if m3 < m1:
            year3 = year3 + 1
        c2 = [x for x in contracts if x.month == m2 and x.year == year2]
        c3 = [x for x in contracts if x.month == m3 and x.year == year3]
        if len(c2) == 1 and len(c3) == 1:
            c2, c3 = c2[0], c3[0]
            s = contracts[c1] + contracts[c3] - (2 * contracts[c2])
            s.name = year1
            dfs.append(s)

    res = pd.concat(dfs, 1)
    res = res.dropna(how='all', axis='rows')
    return res


def fly_quarterly(contracts, x, y, z):
    """
    Given a dataframe of quarterly contract values (eg Brent Q115, Brent Q215, Brent Q315)
    with columns headings as 'Q1 2015', 'Q2 2015'
    Return a dataframe of flys  (eg x = q1 y = q2 z = q3 gives Q1/Q2/Q3 fly)
    """

    dfs = []
    cf = [n for n in contracts if 'Q%s' % x in n]
    for c1 in cf:
        year1, year2, year3 = int(c1[-4:]), int(c1[-4:]), int(c1[-4:])
        # year rollover

        if x == 4 and y == 1:  # 412 or 413
            year2 = year2 + 1
            year3 = year3 + 1
        if (x == 2 and y == 3 and z == 1) or (x == 2 and y == 3 and z == 1):
            year3 = year3 + 1
        if x == 3 and y == 4:  # 341 or 342
            year3 = year3 + 1

        c2 = [n for n in contracts if 'Q%d' % y in n and str(year2) in n]
        c3 = [n for n in contracts if 'Q%d' % z in n and str(year3) in n]
        if len(c2) == 1 and len(c3) == 1:
            c2, c3 = c2[0], c3[0]
            s = contracts[c1] + contracts[c3] - (2 * contracts[c2])
            s.name = 'Q%dQ%dQ%d %d' % (x, y, z, year1)
            dfs.append(s)

    res = pd.concat(dfs, 1)
    res = res.dropna(how='all', axis='rows')
    return res


def time_spreads(contracts, m1, m2):
    """
    Given a dataframe of daily values for monthly contracts (eg Brent Jan 15, Brent Feb 15, Brent Mar 15)
    with columns headings as '2020-01-01', '2020-02-01'
    Return a dataframe of time spreads  (eg m1 = 12, m2 = 12 gives Dec-Dec spread)
    """
    if isinstance(m1, int) and isinstance(m2, int):
        return time_spreads_monthly(contracts, m1, m2)

    if m1.lower().startswith('q') and m2.lower().startswith('q'):
        return time_spreads_quarterly(contracts, m1, m2)


def quarterly_contracts(c):
    """
    Given a dataframe of daily values for monthly contracts (eg Brent Jan 15, Brent Feb 15, Brent Mar 15)
    with columns headings as '2020-01-01', '2020-02-01'
    Return a dataframe of quarterly values (eg Q115)
    """
    years = list(set([x.year for x in c.columns]))

    dfs = []
    for year in years:
        c1, c2, c3 = '{}-01-01'.format(year), '{}-02-01'.format(year), '{}-03-01'.format(year)
        if c1 in c.columns and c2 in c.columns and c3 in c.columns:
            s = pd.concat([c[c1], c[c2], c[c3]], 1).dropna(how='any').mean(axis=1)
            s.name = 'Q1 {}'.format(year)
            dfs.append(s)

        c4, c5, c6 = '{}-04-01'.format(year), '{}-05-01'.format(year), '{}-06-01'.format(year)
        if c4 in c.columns and c5 in c.columns and c6 in c.columns:
            s = pd.concat([c[c4], c[c5], c[c6]], 1, sort=True).dropna(how='any').mean(axis=1)
            s.name = 'Q2 {}'.format(year)
            dfs.append(s)

        c7, c8, c9 = '{}-07-01'.format(year), '{}-08-01'.format(year), '{}-09-01'.format(year)
        if c7 in c.columns and c8 in c.columns and c9 in c.columns:
            s = pd.concat([c[c7], c[c8], c[c9]], 1).dropna(how='any').mean(axis=1)
            s.name = 'Q3 {}'.format(year)
            dfs.append(s)

        c10, c11, c12 = '{}-10-01'.format(year), '{}-11-01'.format(year), '{}-12-01'.format(year)
        if c10 in c.columns and c11 in c.columns and c12 in c.columns:
            s = pd.concat([c[c10], c[c11], c[c12]], 1).dropna(how='any').mean(axis=1)
            s.name = 'Q4 {}'.format(year)
            dfs.append(s)

    res = pd.concat(dfs, 1)
    # sort columns by years
    cols = list(res.columns)
    cols.sort(key=lambda s: s.split()[1])
    res = res[cols]
    return res


def quarterly_spreads(q):
    """
    Given a dataframe of quarterly contract values (eg Brent Q115, Brent Q215, Brent Q315)
    with columns headings as 'Q1 2015', 'Q2 2015'
    Return a dataframe of quarterly spreads (eg Q1-Q2 15)
    Does Q1-Q2, Q2-Q3, Q3-Q4, Q4-Q1
    """
    sprmap = {
        'Q1': 'Q2 {}',
        'Q2': 'Q3 {}',
        'Q3': 'Q4 {}',
        'Q4': 'Q1 {}',
    }

    qtrspr = []
    for col in q.columns:
        colqx = col.split(' ')[0]
        colqxyr = col.split(' ')[1]
        if colqx == 'Q4':
            colqxyr = int(colqxyr) + 1
        colqy = sprmap.get(colqx).format(colqxyr)
        if colqy in q.columns:
            r = q[col] - q[colqy]
            r.name = '{}{} {}'.format(colqx, colqy.split(' ')[0], colqxyr)
            qtrspr.append(r)

    res = pd.concat(qtrspr, 1, sort=True)
    return res


def quarterly_flys(q):
    """
    Given a dataframe of quarterly contract values (eg Brent Q115, Brent Q215, Brent Q315)
    with columns headings as 'Q1 2015', 'Q2 2015'
    Return a dataframe of quarterly flys (eg Q1Q2Q3)
    Does Q1Q2Q3, Q2Q3Q4, Q3Q4Q1, Q4Q1Q2
    """
    flycombos = ((1, 2, 3), (2, 3, 4), (3, 4, 1), (4, 1, 2))

    dfs = []
    for flycombo in flycombos:
        s = fly_quarterly(contracts=q, x=flycombo[0], y=flycombo[1], z=flycombo[2])
        dfs.append(s)

    res = pd.concat(dfs, 1, sort=True)
    return res


def relevant_qtr_contract(qx):
    """
    Given a qtr, eg, Q1, determine the right year to use in seasonal charts.
    For example after Feb 2020, use Q1 2021 as Q1 2020 would have stopped pricing

    :param qx:
    :return:
    """
    relyear = dates.curyear
    if qx == 'Q1':
        if dates.curmon >= 1:
            relyear = relyear + 1
    elif qx == 'Q2':
        if dates.curmon >= 4:
            relyear = relyear + 1
    elif qx == 'Q3':
        if dates.curmon >= 7:
            relyear = relyear + 1
    elif qx == 'Q4':
        if dates.curyear >= 10:
            relyear = relyear + 1

    return relyear


def cal_contracts(c):
    """
    Given a dataframe of daily values for monthly contracts (eg Brent Jan 15, Brent Feb 15, Brent Mar 15)
    with columns headings as '2020-01-01', '2020-02-01'
    Return a dataframe of cal values (eg Cal15)
    """
    years = list(set([x.year for x in c.columns]))

    dfs = []
    for year in years:
        s = c[[x for x in c.columns if x.year == year]].dropna(how='all', axis=1)
        if len(s.columns) == 12:  # only do if we have full set of contracts
            s = s.mean(axis=1)
            s.name = 'CAL {}'.format(year)
            dfs.append(s)
        elif year == dates.curyear and len(
                s.columns) > 0:  # sometimes current year passed in has less than 12 columns but should be included
            s = s.mean(axis=1)
            s.name = 'CAL {}'.format(year)
            dfs.append(s)

    res = pd.concat(dfs, 1)
    # sort columns by years
    cols = list(res.columns)
    cols.sort(key=lambda s: s.split()[1])
    res = res[cols]
    return res


def cal_spreads(q):
    """
    Given a dataframe of cal contract values (eg CAL 2015, CAL 2020)
    with columns headings as 'CAL 2015', 'CAL 2020'
    Return a dataframe of cal spreads (eg CAL 2015-2016)
    """

    calspr = []
    for col in q.columns:
        # colcal = col.split(' ')[0]
        colcalyr = col.split(' ')[1]

        curyear = int(colcalyr)
        nextyear = curyear + 1

        colcalnextyr = 'CAL %s' % (nextyear)
        if colcalnextyr in q.columns:
            r = q[col] - q[colcalnextyr]
            r.name = 'CAL {}-{}'.format(curyear, nextyear)
            calspr.append(r)

    if len(calspr) > 0:
        res = pd.concat(calspr, 1, sort=True)
        return res


def spread_combinations(contracts):
    output = {}
    output['Calendar'] = cal_contracts(contracts)
    output['Calendar Spread'] = cal_spreads(output['Calendar'])
    output['Quarterly'] = quarterly_contracts(contracts)

    q = output['Quarterly']
    for qx in ['Q1', 'Q2', 'Q3', 'Q4']:
        output[qx] = q[[x for x in q if qx in x]]
    output['Quarterly Spread'] = quarterly_spreads(q)
    q = output['Quarterly Spread']
    for qx in ['Q1-Q2', 'Q2-Q3', 'Q3-Q4', 'Q4-Q1']:
        output[qx] = q[[x for x in q if qx in x]]

    for month in [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]:
        output[month] = contracts[[x for x in contracts.columns if x.month == month]]

    for spread in [[1, 2], [2, 3], [3, 4], [4, 5], [5, 6], [6, 7], [7, 8], [8, 9], [9, 10], [10, 11], [11, 12], [12, 1],
                   [6, 6], [6, 12], [12, 12], [10, 12], [4, 9], [10, 3]]:
        tag = '%s%s' % (month_abbr[spread[0]], month_abbr[spread[1]])
        output[tag] = time_spreads(contracts, spread[0], spread[1])

    for flyx in [[1, 2, 3], [2, 3, 4], [3, 4, 5], [4, 5, 6], [5, 6, 7], [6, 7, 8], [7, 8, 9], [8, 9, 10], [9, 10, 11],
                 [10, 11, 12], [11, 12, 1], [12, 1, 2]]:
        tag = '%s%s%s' % (month_abbr[flyx[0]], month_abbr[flyx[1]], month_abbr[flyx[2]])
        output[tag] = fly(contracts, flyx[0], flyx[1], flyx[2])

    return output


def spread_combination(contracts, combination_type, verbose_columns=True):
    """
    Convenience method to access functionality in forwards using a combination_type keyword
    :param contracts:
    :param combination_type:
    :return:
    """
    combination_type = combination_type.lower()
    contracts = contracts.dropna(how='all', axis='rows')

    if combination_type == "calendar":
        c_contracts = cal_contracts(contracts)
        colmap = dates.find_year(c_contracts)
        c_contracts = c_contracts.rename(columns={x: colmap[x] for x in c_contracts.columns})
        return c_contracts
    if combination_type == "calendar spread":
        c_contracts = cal_spreads(cal_contracts(contracts))
        if not verbose_columns:
            colmap = dates.find_year(c_contracts)
            c_contracts = c_contracts.rename(columns={x: colmap[x] for x in c_contracts.columns})
        return c_contracts
    if combination_type.startswith('q'):
        q_contracts = quarterly_contracts(contracts)
        m = re.search('q\dq\dq\d', combination_type)
        if m:
            q_spreads = fly_quarterly(q_contracts, x=int(combination_type[1]), y=int(combination_type[3]),
                                      z=int(combination_type[5]))
            if not verbose_columns:
                colmap = dates.find_year(q_spreads)
                q_spreads = q_spreads.rename(columns={x: colmap[x] for x in q_spreads.columns})
            return q_spreads
        m = re.search('q\dq\d', combination_type)
        if m:
            q_spreads = quarterly_spreads(q_contracts)
            q_spreads = q_spreads[[x for x in q_spreads.columns if x.startswith(combination_type.upper())]]
            if not verbose_columns:
                colmap = dates.find_year(q_spreads)
                q_spreads = q_spreads.rename(columns={x: colmap[x] for x in q_spreads.columns})
            return q_spreads

        m = re.search('q\d', combination_type)
        if m:
            q_contracts = q_contracts[[x for x in q_contracts.columns if x.startswith(combination_type.upper())]]
            if not verbose_columns:
                colmap = dates.find_year(q_contracts)
                q_contracts = q_contracts.rename(columns={x: colmap[x] for x in q_contracts.columns})
            return q_contracts

    # handle monthly, spread and fly inputs
    month_abbr_inv = {month.lower(): index for index, month in enumerate(month_abbr) if month}
    months = [x.lower() for x in month_abbr]
    if len(combination_type) == 3 and combination_type in months:
        c = contracts[[x for x in contracts if x.month == month_abbr_inv[combination_type]]]
        if verbose_columns:
            c = c.rename(columns={x: x.strftime('%b %Y') for x in c.columns})
        else:
            c = c.rename(columns={x: x.year for x in c.columns})
        return c
    if len(combination_type) == 6:
        m1, m2 = combination_type[0:3], combination_type[3:6]
        if m1 in months and m2 in months:
            c = time_spreads(contracts, month_abbr_inv[m1], month_abbr_inv[m2])
            if verbose_columns:
                c = c.rename(columns={x: '%s%s %s' % (m1.title(), m2.title(), x) for x in c.columns})
            return c
    if len(combination_type) == 9:
        m1, m2, m3 = combination_type[0:3], combination_type[3:6], combination_type[6:9]
        if m1 in months and m2 in months and m3 in months:
            c = fly(contracts, month_abbr_inv[m1], month_abbr_inv[m2], month_abbr_inv[m3])
            if verbose_columns:
                c = c.rename(columns={x: '%s%s%s %s' % (m1.title(), m2.title(), m3.title(), x) for x in c.columns})
            return c



from commodutil import convfactors
import unittest
import pandas as pd


class TestUtils(unittest.TestCase):

    def test_conv_factor(self):
        res = convfactors.convfactor('diesel', 'kt', 'km3')
        self.assertAlmostEqual(res, 1.186, 3)

        res = convfactors.convfactor('ulsd', 'kt', 'km3')
        self.assertAlmostEqual(res, 1.186, 3)

        res = convfactors.convfactor('ulsd', 'km3', 'kt')
        self.assertAlmostEqual(res, 0.843, 3)

    def test_convert(self):
        diesel_kt = 520

        res = convfactors.convert(diesel_kt, 'diesel')
        self.assertAlmostEqual(res, 520, 3)

        res = convfactors.convert(diesel_kt, 'diesel', 'kt', 'km3')
        self.assertAlmostEqual(res, 616.84, 2)

    def test_convert_2(self):
        d = pd.Series([50, 60, 70, 80], index=pd.date_range('2020', periods=4, freq='MS'))
        res = convfactors.convert(d, 'diesel', 'kt', 'bbl/d')
        self.assertAlmostEqual(res['2020-1'].iloc[0], 12.02, 2)

        res = convfactors.convert(d, 'diesel', 'bbl/d', 'kt')
        self.assertAlmostEqual(res['2020-1'].iloc[0], 208.05, 2)

    def test_ec(self):

        diesel_kt = 1
        res = convfactors.convert(diesel_kt, 'diesel', 'kt', 'gj')
        self.assertAlmostEqual(res, 42.7, 2)


if __name__ == '__main__':
    unittest.main()


from commodutil import transforms
from commodutil import dates
import re
import pandas as pd


def curve_seasonal_zscore(hist, fwd):
    """
    Given some history for a timeseries and a forward curve, calculate the monthly
    z-score (std dev away from mean) along the forward curve
    """

    d = transforms.monthly_mean(hist).T.describe()

    if isinstance(fwd, pd.Series):
        fwd = pd.DataFrame(fwd)
    fwd['zscore'] = fwd.apply(lambda x: (d[x.name.month].loc['mean'] - x.iloc[0]) / d[x.name.month].loc['std'], 1)
    return fwd


def reindex_zscore(df, range=10):
    """
    Given a dataframe of contracts (or spreads), calculate z-score for current year onwards
    Essentially returns how far away the 'curve' is from historical trading range
    """
    df = transforms.reindex_year(df)
    df = df.rename(columns={x: int(re.findall('\d\d\d\d', str(x))[0]) for x in df.columns})  # turn columns into years
    d = df.loc[:, dates.curyear - range - 1:dates.curyear - 1]  # get subset of range years
    d = d[:-10] # exclude last 10 rows to due to volatility close to expire

    dfs = []
    for year in df.loc[:, dates.curyear:df.columns[-1]]:
        z = (d.mean(axis=1) - df.loc[:, year]) / d.std(axis=1)
        z.name = year
        dfs.append(z)
    res = pd.concat(dfs, axis=1)

    return res



import pandas as pd
import numpy as np
import re


def mergets(left, right, leftl=None, rightl=None, how='left'):
    """
    Wrapper for pandas merge for working on merging timeseries
    """

    if isinstance(left, pd.Series):
        left = pd.DataFrame(left)
    if isinstance(right, pd.Series):
        right = pd.DataFrame(right)
    res = pd.merge(left, right, left_index=True, right_index=True, how=how)

    rename = {}
    if leftl is not None:
        rename[left.columns[0]] = leftl
        rename['{}_x'.format(left.columns[0])] = leftl
    if rightl is not None:
        rename[right.columns[0]] = rightl
        rename['{}_y'.format(right.columns[0])] = rightl

    res = res.rename(columns=rename)

    return res


def fillna_downbet(df):
    """
    Fill weekends/holidays in timeseries but dont extend to NaNs at end of the timeseries
    https://stackoverflow.com/questions/28136663/using-pandas-to-fill-gaps-only-and-not-nans-on-the-ends
    :param df:
    :return:
    """
    df = df.copy()
    for col in df.columns:
        non_nans = df[col][~df[col].apply(np.isnan)]
        if non_nans is not None and len(non_nans) > 1:
            start, end = non_nans.index[0], non_nans.index[-1]
            df[col].loc[start:end] = df[col].loc[start:end].fillna(method='ffill')
    return df


def sql_insert_statement_from_dataframe(df, table_name, print_statemnt=False):
    """
    Turn a dataframe into a set of insert statements
    Taken from https://stackoverflow.com/questions/31071952/generate-sql-statements-from-a-pandas-dataframe
    :param df:
    :param table_name:
    :return:
    """
    sql_texts = []
    for index, row in df.iterrows():
        vals = [re.sub(r'\'', "", x) if isinstance(x, str) else x for x in row.values]
        q = 'INSERT INTO ' + table_name + ' (' + str(', '.join(df.columns)) + ') VALUES ' + str(tuple(vals))
        q = q.replace('nan', 'Null').replace('None', 'Null')
        if print_statemnt:
            print(q)
        sql_texts.append(q)

    return sql_texts



from commodutil import dates
import unittest
import pandas as pd


class TestDates(unittest.TestCase):

    def test_find_year(self):
        df = pd.DataFrame(columns=['Q1 2020', 'Q2 2022'])
        res = dates.find_year(df)
        self.assertEqual(res['Q1 2020'], 2020)
        self.assertEqual(res['Q2 2022'], 2022)

    def test_find_year2(self):
        df = pd.DataFrame(columns=['CAL 2020-2021'])
        res = dates.find_year(df)
        self.assertEqual(res['CAL 2020-2021'], 2020)

    def test_find_year3(self):
        df = pd.DataFrame(columns=['FB', 'FP'])
        res = dates.find_year(df)
        self.assertEqual(res['FB'], 'FB')
        self.assertEqual(res['FP'], 'FP')

    def test_find_year4(self):
        df = pd.DataFrame(columns=['FB', 'FP 2021'])
        res = dates.find_year(df)
        self.assertEqual(res['FB'], 'FB')
        self.assertEqual(res['FP 2021'], 2021)


if __name__ == '__main__':
    unittest.main()


from commodutil import forwards
import unittest
import os
import pandas as pd


class TestForwards(unittest.TestCase):

    def test_conv_factor(self):
        res = forwards.convert_contract_to_date('2020F')
        self.assertEqual(res, '2020-1-1')

    def test_quarterly_contracts(self):
        dirname, filename = os.path.split(os.path.abspath(__file__))
        cl = pd.read_csv(os.path.join(dirname, 'test_cl.csv'), index_col=0, parse_dates=True, dayfirst=True)
        contracts = cl.rename(columns={x: pd.to_datetime(forwards.convert_contract_to_date(x)) for x in cl.columns})

        res = forwards.quarterly_contracts(contracts)
        self.assertAlmostEqual(res['Q2 2019'].loc[pd.to_datetime('2019-03-20')], 60.18, 2)
        self.assertAlmostEqual(res['Q3 2019'].loc[pd.to_datetime('2019-06-20')], 56.95, 2)
        self.assertAlmostEqual(res['Q4 2019'].loc[pd.to_datetime('2019-09-20')], 58.01, 2)
        self.assertAlmostEqual(res['Q1 2020'].loc[pd.to_datetime('2019-12-19')], 61.09, 2)

        self.assertAlmostEqual(res['Q2 2020'].loc[pd.to_datetime('2020-03-20')], 23.14, 2)

        res_qs = forwards.quarterly_spreads(res)
        self.assertAlmostEqual(res_qs['Q1Q2 2020'].loc[pd.to_datetime('2019-12-19')], 1.14, 2)
        self.assertAlmostEqual(res_qs['Q2Q3 2019'].loc[pd.to_datetime('2019-03-20')], -0.73, 2)
        self.assertAlmostEqual(res_qs['Q3Q4 2019'].loc[pd.to_datetime('2019-06-20')], 0.07, 2)
        self.assertAlmostEqual(res_qs['Q4Q1 2020'].loc[pd.to_datetime('2019-09-20')], 1.12, 2)

        res_qf = forwards.quarterly_flys(res)
        self.assertAlmostEqual(res_qf['Q1Q2Q3 2020'].loc[pd.to_datetime('2019-12-19')], -0.53, 2)
        self.assertAlmostEqual(res_qf['Q2Q3Q4 2019'].loc[pd.to_datetime('2019-03-20')], -0.66, 2)
        self.assertAlmostEqual(res_qf['Q3Q4Q1 2019'].loc[pd.to_datetime('2019-06-20')], -0.58, 2)
        self.assertAlmostEqual(res_qf['Q4Q1Q2 2020'].loc[pd.to_datetime('2019-09-20')], 0.21, 2)

    def test_cal_contracts(self):
        dirname, filename = os.path.split(os.path.abspath(__file__))
        cl = pd.read_csv(os.path.join(dirname, 'test_cl.csv'), index_col=0, parse_dates=True, dayfirst=True)
        contracts = cl.rename(columns={x: pd.to_datetime(forwards.convert_contract_to_date(x)) for x in cl.columns})

        res = forwards.cal_contracts(contracts)
        self.assertAlmostEqual(res['CAL 2020'].loc[pd.to_datetime('2019-03-20')], 59.53, 2)
        self.assertAlmostEqual(res['CAL 2021'].loc[pd.to_datetime('2019-03-20')], 57.19, 2)

        res = forwards.cal_spreads(res)
        self.assertAlmostEqual(res['CAL 2020-2021'].loc[pd.to_datetime('2019-12-19')], 4.77, 2)
        self.assertAlmostEqual(res['CAL 2021-2022'].loc[pd.to_datetime('2019-03-20')], 1.77, 2)

    def test_timespreads(self):
        dirname, filename = os.path.split(os.path.abspath(__file__))
        cl = pd.read_csv(os.path.join(dirname, 'test_cl.csv'), index_col=0, parse_dates=True, dayfirst=True)
        contracts = cl.rename(columns={x: pd.to_datetime(forwards.convert_contract_to_date(x)) for x in cl.columns})

        res = forwards.time_spreads(contracts, m1=6, m2=12)
        self.assertAlmostEqual(res[2019].loc[pd.to_datetime('2019-01-02')], -1.51, 2)
        self.assertAlmostEqual(res[2019].loc[pd.to_datetime('2019-05-21')], 0.37, 2)

        res = forwards.time_spreads(contracts, m1=12, m2=12)
        self.assertAlmostEqual(res[2019].loc[pd.to_datetime('2019-11-20')], 3.56, 2)
        self.assertAlmostEqual(res[2020].loc[pd.to_datetime('2019-03-20')], 2.11, 2)

    def test_timespreads2(self):
        dirname, filename = os.path.split(os.path.abspath(__file__))
        cl = pd.read_csv(os.path.join(dirname, 'test_cl.csv'), index_col=0, parse_dates=True, dayfirst=True)
        contracts = cl.rename(columns={x: pd.to_datetime(forwards.convert_contract_to_date(x)) for x in cl.columns})

        res = forwards.time_spreads(contracts, m1='Q1', m2='Q2')
        self.assertAlmostEqual(res[2020].loc[pd.to_datetime('2019-01-02')], -0.33, 2)
        self.assertAlmostEqual(res[2020].loc[pd.to_datetime('2019-05-21')], 1.05, 2)

        res = forwards.time_spreads(contracts, m1='Q4', m2='Q1')
        self.assertAlmostEqual(res[2020].loc[pd.to_datetime('2019-01-02')], -0.25, 2)
        self.assertAlmostEqual(res[2020].loc[pd.to_datetime('2019-05-21')], 0.91, 2)

    def test_fly(self):
        dirname, filename = os.path.split(os.path.abspath(__file__))
        cl = pd.read_csv(os.path.join(dirname, 'test_cl.csv'), index_col=0, parse_dates=True, dayfirst=True)
        contracts = cl.rename(columns={x: pd.to_datetime(forwards.convert_contract_to_date(x)) for x in cl.columns})

        res = forwards.fly(contracts, m1=1, m2=2, m3=3)
        self.assertAlmostEqual(res[2020].loc[pd.to_datetime('2019-01-03')], -0.02, 2)
        self.assertAlmostEqual(res[2021].loc[pd.to_datetime('2019-05-21')], 0.02, 2)

    def test_fly2(self):
        dirname, filename = os.path.split(os.path.abspath(__file__))
        cl = pd.read_csv(os.path.join(dirname, 'test_cl.csv'), index_col=0, parse_dates=True, dayfirst=True)
        contracts = cl.rename(columns={x: pd.to_datetime(forwards.convert_contract_to_date(x)) for x in cl.columns})

        res = forwards.fly(contracts, m1=12, m2=1, m3=3)
        self.assertAlmostEqual(res[2020].loc[pd.to_datetime('2019-01-03')], 0.06, 2)
        self.assertAlmostEqual(res[2021].loc[pd.to_datetime('2019-05-21')], -0.14, 2)

    def test_fly_quarterly(self):
        dirname, filename = os.path.split(os.path.abspath(__file__))
        cl = pd.read_csv(os.path.join(dirname, 'test_cl.csv'), index_col=0, parse_dates=True, dayfirst=True)
        contracts = cl.rename(columns={x: pd.to_datetime(forwards.convert_contract_to_date(x)) for x in cl.columns})
        contracts = forwards.quarterly_contracts(contracts)
        res = forwards.fly_quarterly(contracts, x=1, y=2, z=3)
        self.assertAlmostEqual(res['Q1Q2Q3 2020'].loc[pd.to_datetime('2019-01-03')], -0.073, 3)
        self.assertAlmostEqual(res['Q1Q2Q3 2021'].loc[pd.to_datetime('2019-05-21')], 0.11, 2)

    def test_spread_combinations(self):
        dirname, filename = os.path.split(os.path.abspath(__file__))
        cl = pd.read_csv(os.path.join(dirname, 'test_cl.csv'), index_col=0, parse_dates=True, dayfirst=True)
        contracts = cl.rename(columns={x: pd.to_datetime(forwards.convert_contract_to_date(x)) for x in cl.columns})

        res = forwards.spread_combinations(contracts)
        self.assertIn('Q1', res)
        self.assertIn('Q1-Q2', res)
        self.assertIn('Calendar', res)
        self.assertIn('JanFeb', res)
        self.assertIn('JanFebMar', res)

    def test_spread_combination_calendar(self):
        dirname, filename = os.path.split(os.path.abspath(__file__))
        cl = pd.read_csv(os.path.join(dirname, 'test_cl.csv'), index_col=0, parse_dates=True, dayfirst=True)
        contracts = cl.rename(columns={x: pd.to_datetime(forwards.convert_contract_to_date(x)) for x in cl.columns})

        res = forwards.spread_combination(contracts, 'calendar')
        self.assertIsNotNone(res)
        self.assertAlmostEqual(res[2020]['2020-01-02'], 59.174, 3)

    def test_spread_combination_calendar_spread(self):
        dirname, filename = os.path.split(os.path.abspath(__file__))
        cl = pd.read_csv(os.path.join(dirname, 'test_cl.csv'), index_col=0, parse_dates=True, dayfirst=True)
        contracts = cl.rename(columns={x: pd.to_datetime(forwards.convert_contract_to_date(x)) for x in cl.columns})

        res = forwards.spread_combination(contracts, 'calendar spread')
        self.assertAlmostEqual(res['CAL 2020-2021']['2020-01-02'], 4.35, 2)

    def test_spread_combination_quarter(self):
        dirname, filename = os.path.split(os.path.abspath(__file__))
        cl = pd.read_csv(os.path.join(dirname, 'test_cl.csv'), index_col=0, parse_dates=True, dayfirst=True)
        contracts = cl.rename(columns={x: pd.to_datetime(forwards.convert_contract_to_date(x)) for x in cl.columns})

        res = forwards.spread_combination(contracts, 'q1')
        self.assertAlmostEqual(res['Q1 2020']['2019-01-02'], 49.88, 2)

    def test_spread_combination_quarter_spread(self):
        dirname, filename = os.path.split(os.path.abspath(__file__))
        cl = pd.read_csv(os.path.join(dirname, 'test_cl.csv'), index_col=0, parse_dates=True, dayfirst=True)
        contracts = cl.rename(columns={x: pd.to_datetime(forwards.convert_contract_to_date(x)) for x in cl.columns})

        res = forwards.spread_combination(contracts, 'q1q2')
        self.assertAlmostEqual(res['Q1Q2 2020']['2019-01-02'], -0.33, 2)

    def test_spread_combination_month(self):
        dirname, filename = os.path.split(os.path.abspath(__file__))
        cl = pd.read_csv(os.path.join(dirname, 'test_cl.csv'), index_col=0, parse_dates=True, dayfirst=True)
        contracts = cl.rename(columns={x: pd.to_datetime(forwards.convert_contract_to_date(x)) for x in cl.columns})

        res = forwards.spread_combination(contracts, 'jan')
        self.assertAlmostEqual(res['Jan 2020']['2019-01-02'], 49.77, 2)

    def test_spread_combination_month_spread(self):
        dirname, filename = os.path.split(os.path.abspath(__file__))
        cl = pd.read_csv(os.path.join(dirname, 'test_cl.csv'), index_col=0, parse_dates=True, dayfirst=True)
        contracts = cl.rename(columns={x: pd.to_datetime(forwards.convert_contract_to_date(x)) for x in cl.columns})

        res = forwards.spread_combination(contracts, 'janfeb')
        self.assertAlmostEqual(res['JanFeb 2020']['2019-01-02'], -0.11, 2)

    def test_spread_combination_month_fly(self):
        dirname, filename = os.path.split(os.path.abspath(__file__))
        cl = pd.read_csv(os.path.join(dirname, 'test_cl.csv'), index_col=0, parse_dates=True, dayfirst=True)
        contracts = cl.rename(columns={x: pd.to_datetime(forwards.convert_contract_to_date(x)) for x in cl.columns})

        res = forwards.spread_combination(contracts, 'janfebmar')
        self.assertAlmostEqual(res['JanFebMar 2020']['2019-01-02'], 0.0, 2)

    def test_spread_combination_quarter_spread(self):
        dirname, filename = os.path.split(os.path.abspath(__file__))
        cl = pd.read_csv(os.path.join(dirname, 'test_cl.csv'), index_col=0, parse_dates=True, dayfirst=True)
        contracts = cl.rename(columns={x: pd.to_datetime(forwards.convert_contract_to_date(x)) for x in cl.columns})

        res = forwards.spread_combination(contracts, 'q4q1q2')
        self.assertAlmostEqual(res['Q4Q1Q2 2020']['2019-01-02'], -0.023, 3)


if __name__ == '__main__':
    unittest.main()


import unittest
import pandas as pd
from commodutil import pandasutil
from commodutil import forwards
import os


class TestPandasUtils(unittest.TestCase):

    def test_mergets(self):
        dirname, filename = os.path.split(os.path.abspath(__file__))
        cl = pd.read_csv(os.path.join(dirname, 'test_cl.csv'), index_col=0, parse_dates=True, dayfirst=True)
        contracts = cl.rename(columns={x: pd.to_datetime(forwards.convert_contract_to_date(x)) for x in cl.columns})

        res = pandasutil.mergets(contracts['2020-01-01'], contracts['2020-02-01'], leftl='Test1', rightl='Test2')
        self.assertIn('Test1', res.columns)
        self.assertIn('Test2', res.columns)

    def test_sql_insert(self):
        df = pd.DataFrame([[1,2,3], [4,'test\'ing',6], [7,8,9]], columns=['a', 'b', 'c'])
        res = pandasutil.sql_insert_statement_from_dataframe(df, 'table')
        exp = 'INSERT INTO table (a, b, c) VALUES (1, 2, 3)'
        self.assertEqual(res[0], exp)
        exp = "INSERT INTO table (a, b, c) VALUES (4, 'testing', 6)"
        self.assertEqual(res[1], exp)


if __name__ == '__main__':
    unittest.main()





####
# Commodity conversion factors
#
# - Convention:
#       commodity_from_to
#
# eg: diesel_kt_bbl
#
# optionally specify parameters for non standard entities: eg diesel_kt_bbl_iea
#
# standard should be to multiply the conversion factor
# eg 1 mt * 7.45 = 7.45 bbls
#
# bbl = barrels
# km3 = cubmic meters
# kt = kilo atonne
# ltr = litre
# gj = 1000 mj # thus km3_gj is same as l_mj


import pandas as pd


# converts different names for commodities to standard used in this file
commonnamemap = {
    'dies': 'diesel',
    'ulsd': 'diesel',
    'go': 'diesel',
    'gas': 'gasoline', # tricky - could be NatGag
    'etho': 'ethanol',

}

# API
api_modulus = 141.5

# bbls to km3
km3_bbl = 6.28981

# Gallons to litres
# 1 US gallon = 0.83267384 Imp gallon = 3.78541178 Liters
gal_ltr = 3.78541178


# Light Ends
naphtha_kt_bbl = 8.9

gasoline_density_kg_l = 0.745
gasoline_kt_bbl = 8.33
gasoline_kt_km3 = 1/gasoline_density_kg_l  # 1.342281879
gasoline_km3_bbl = gasoline_kt_bbl / gasoline_kt_km3
gasoline_gal_kt = gal_ltr * (gasoline_density_kg_l / 1000)
gasoline_km3_gj = 32


ccs_kt_bbl = 8.62
isomerate_kt_bbl = 9.53
alkylate_kt_bbl = 8.99
reformate_kt_bbl = 7.07


ethanol_density_kg_l = 0.789
ethanol_kt_bbl = 8.33  #7.96
ethanol_kt_km3 = 1/ethanol_density_kg_l # 1.2674
ethanol_km3_bbl = ethanol_kt_bbl / ethanol_kt_km3
ethanol_gal_kt = gal_ltr * (ethanol_density_kg_l / 1000)
ethanol_km3_gj = 21


etbe_density_kg_l = 0.745
etbe_kt_km3 = 1/ethanol_density_kg_l # 1.34
etbe_km3_gj = 27.04

# Middle distillates
jet_kt_bbl = 7.88

diesel_density_kg_l = 0.843
diesel_kt_bbl = 7.45
diesel_kt_km3 = 1/diesel_density_kg_l  # 1.18623962
diesel_km3_bbl = diesel_kt_bbl / diesel_kt_km3
diesel_gal_kt = gal_ltr * (diesel_density_kg_l / 1000)
diesel_km3_gj = 36


fame_density_kg_l = 0.892
fame_kt_bbl = 7.051345
fame_kt_km3 = 1/ fame_density_kg_l  # 1.121076233
fame_km3_bbl = fame_kt_bbl / diesel_kt_km3
fame_gal_kt = gal_ltr * (fame_density_kg_l / 1000)
fame_km3_gj = 33


b7_density_kg_l = 0.845
b7_kt_bbl = 7.078
b7_kt_km3 = 1/ b7_density_kg_l  # 1.11834
b7_km3_bbl = b7_kt_bbl / diesel_kt_km3
b7_gal_kt = gal_ltr * (b7_density_kg_l / 1000)
b7_km3_gj = 35.8


hvo_density_kg_l = 0.778
hvo_kt_bbl = 8.046
hvo_kt_km3 = 1/hvo_density_kg_l  # 1.285347044
hvo_km3_bbl = hvo_kt_bbl / hvo_kt_km3
hvo_gal_kt = gal_ltr * (hvo_density_kg_l / 1000)
hvo_km3_gj = 34


# FO / Heave
vgo_kt_bbl = 6.9
sr_fo_kt_bbl = 6.65
fo_kt_bbl = 6.35


# Crude


# Natgas

natgas_density_kg_l = 0.542
natgas_kt_bbl = 11.6
natgas_kt_km3 = 1/natgas_density_kg_l  # 1.844
natgas_km3_bbl = natgas_kt_bbl / natgas_kt_km3
natgas_gal_kt = gal_ltr * (natgas_density_kg_l / 1000)
natgas_km3_gj = 26.137 #https://callmepower.ca/en/faq/gigajoule-cubic-metre-gas


# print diesel_km3_bbl
# print fame_gal_mt
# print gasoline_gal_kt
# print diesel_gal_kt

def _stdcmmd(cmmdstring):
    global commonnamemap
    cmmdstring = cmmdstring.lower()
    cmmdstring = commonnamemap.get(cmmdstring, cmmdstring)
    return cmmdstring


def _stdunits(unit):
    if unit == 'mt':
        return 'kt'
    if unit == 'm3':
        return 'km3'

    return unit


def convfactor(commodity, fromunit, tounit):
    commodity = _stdcmmd(commodity)

    # TODO - move to mt/m3 convetion rather than kt/km3
    fromunit = _stdunits(fromunit)
    tounit = _stdunits(tounit)

    factor = '{}_{}_{}'.format(commodity, fromunit, tounit)
    if factor in globals():
        return globals()[factor]

    factor = '{}_{}_{}'.format(commodity, tounit, fromunit)
    if factor in globals():
        res = globals()[factor]
        res = 1/res
        return res

    # if af this point we don't have a conversion factor try to mix an match on kt/km3 combinations
    if fromunit == 'kt':
        int_cv = convfactor(commodity, 'km3', tounit)
        kt_km3_cv = convfactor(commodity, 'kt', 'km3')
        new_cv = int_cv * kt_km3_cv
        return new_cv

    if tounit == 'kt':
        int_cv = convfactor(commodity, 'km3', fromunit)
        kt_km3_cv = convfactor(commodity, 'kt', 'km3')
        new_cv = 1/(int_cv * kt_km3_cv)
        return new_cv

    raise ValueError('no conversion factor for {} from: {} to: {}'.format(commodity, fromunit, tounit))


def convert_days_in_month(s, transform='mul'):
    m = pd.Series(s.index.days_in_month, index=s.index)
    if transform == 'mul':
        s = s.mul(m, 0)
    elif transform == 'div':
        s = s.div(m, 0)
    return s


def convert(val, commodity, fromunit=None, tounit=None):
    if fromunit != tounit:
        assert commodity is not None, 'need type of commodity to convert {} from {} to {}'.format(val, fromunit, tounit)
        fromunit_i, tounit_i = fromunit, tounit
        if fromunit.endswith('/d'):
            fromunit_i = fromunit_i.replace('/d', '')
        if tounit.endswith('/d'):
            tounit_i = tounit_i.replace('/d', '')
        unitconvfactor = convfactor(commodity, fromunit_i, tounit_i)

        if fromunit.endswith('/d'):
            val = convert_days_in_month(val, 'mul')

        val = val * unitconvfactor

        if tounit.endswith('/d'):
            val = convert_days_in_month(val, 'div')

    return val


if __name__ == '__main__':
    a = convfactor('diesel', 'kt', 'km3')
    print(a)


import unittest, os
import pandas as pd
from datetime import datetime
from commodutil import forwards
from commodutil import transforms
from commodutil import dates


class TestTransforms(unittest.TestCase):

    def test_seasonalise2(self):
        dirname, filename = os.path.split(os.path.abspath(__file__))
        cl = pd.read_csv(os.path.join(dirname, 'test_cl.csv'), index_col=0, parse_dates=True, dayfirst=True)

        seas = transforms.seasonailse(cl['CL_2020J'], fillna=True)
        self.assertEqual(seas[2020].loc[pd.to_datetime('%s-03-13' % dates.curyear)], 31.73)
        self.assertEqual(seas[2020].loc[pd.to_datetime('%s-03-16' % dates.curyear)], 28.7)
        self.assertEqual(seas[2020].loc[pd.to_datetime('%s-03-17' % dates.curyear)], 26.95)
        self.assertEqual(seas[2020].loc[pd.to_datetime('%s-03-18' % dates.curyear)], 20.37)
        self.assertEqual(seas[2020].loc[pd.to_datetime('%s-03-19' % dates.curyear)], 25.22)

    def test_seasonalise_weekly(self):
        dirname, filename = os.path.split(os.path.abspath(__file__))
        cl = pd.read_csv(os.path.join(dirname, 'test_weekly.csv'), index_col=0, parse_dates=True, dayfirst=True)

        seas = transforms.seasonalise_weekly(cl['PET.WCRSTUS1.W'])

        self.assertEqual(seas[2020].loc[datetime.fromisocalendar(dates.curyear, pd.to_datetime('2020-01-03').isocalendar()[1], 1)], 1066027)
        self.assertEqual(seas[2020].loc[datetime.fromisocalendar(dates.curyear, pd.to_datetime('2020-01-10').isocalendar()[1], 1)], 1063478)
        self.assertEqual(seas[2000].loc[datetime.fromisocalendar(dates.curyear, pd.to_datetime('2000-01-07').isocalendar()[1], 1)], 844791)
        self.assertEqual(seas[2000].loc[datetime.fromisocalendar(dates.curyear, pd.to_datetime('2000-12-29').isocalendar()[1], 1)], 813959)

    def test_reindex_year(self):
        """
        Test reindex with cal spread (eg Cal 20-21, Cal 21-22)
        :return:
        """
        dirname, filename = os.path.split(os.path.abspath(__file__))
        cl = pd.read_csv(os.path.join(dirname, 'test_cl.csv'), index_col=0, parse_dates=True, dayfirst=True)
        contracts = cl.rename(columns={x: pd.to_datetime(forwards.convert_contract_to_date(x)) for x in cl.columns})
        qorig = forwards.cal_contracts(contracts)
        cal_sp = forwards.cal_spreads(qorig)

        res = transforms.reindex_year(cal_sp)
        self.assertAlmostEqual(-1.04, res['CAL 2020-2021']['%s-01-02' % (dates.curyear - 1)], 2)
        self.assertAlmostEqual(2.32, res['CAL 2021-2022']['%s-01-02' % (dates.curyear - 1)], 2)

    def test_monthly_mean(self):
        dirname, filename = os.path.split(os.path.abspath(__file__))
        df = pd.read_csv(os.path.join(dirname, 'test_cl.csv'), index_col=0, parse_dates=True, dayfirst=True)[['CL_2020F']].dropna()

        res = transforms.monthly_mean(df)
        self.assertAlmostEqual(res['CL_2020F'][2015][1], 66.77, 2)


if __name__ == '__main__':
    unittest.main()





import pandas as pd
from datetime import datetime
from functools import reduce
from commodutil import dates
from commodutil import pandasutil


def seasonailse(df, fillna=True):
    if isinstance(df, pd.DataFrame):
        df = pd.Series(df[df.columns[0]])

    assert isinstance(df, pd.Series)

    s = df[~((df.index.month == 2) & (df.index.day == 29))]  # remove leap dates 29 Feb
    seas = s.groupby([s.index.month, s.index.day, s.index.year, ]).mean().unstack()

    # replace index with dates from current year
    newind = [pd.to_datetime('{}-{}-{}'.format(dates.curyear, i[0], i[1])) for i in seas.index]
    seas.index = newind

    if fillna:
        seas = pandasutil.fillna_downbet(seas)

    return seas


def seasonalise_weekly(df, freq='W'):
    """
    Edge case for handling weekly data - eg DOE where we need to tweak the standard
    seasonalise() method.
    :param df:
    :return:
    """
    if isinstance(df, pd.Series):
        df = pd.DataFrame(df)

    df = pd.merge(df, df.index.isocalendar(), left_index=True, right_index=True)
    df = df.groupby([df.week, df.year]).mean()[df.columns[0]].unstack()
    # when converting back to date format, some years don't have week 53 so drop for now
    if 53 in df.index and pd.to_datetime('%s-12-31' % dates.curyear).isocalendar()[1] == 52:
        df = df.drop(53)
    df.index = df.index.map(lambda x: datetime.fromisocalendar(2021, x, 1))
    return df


def forward_only(df):
    """
    Only take forward timeseries from cur month onwards (discarding the history)
    """
    df = df[dates.curmonyear_str:]
    return df


def format_fwd(df, last_index=None):
    """
    Format a monthly-frequency forward curve into a daily series
    """
    df = df.resample('D').mean().fillna(method='ffill')
    if last_index is not None:
        df = df[last_index:]

    return df


def reindex_year(df):
    """
    Reindex a dataframe containing prices to the current year.
    eg dataframe with brent Jan 19, Jan 18, Jan 17   so that 18 is shifted +1 year and 17 is shifted +2 years
    """
    dfs = []
    colyearmap = dates.find_year(df)
    for colname in df.columns:
        if df[colname].isnull().all():
            continue # logic below wont work on all empty NaN columns

        # determine year
        colyear = colyearmap[colname]
        delta = dates.curyear - colyear
        w = df[[colname]]
        if delta == 0:
            dfs.append(w)
        else: # reindex
            winew = [x + pd.DateOffset(years=delta) for x in w.index]
            w.index = winew
            dfs.append(w)

    # merge all series into one dataframe, concat doesn't quite do the job
    res = reduce(lambda left, right: pd.merge(left, right, left_index=True, right_index=True, how='outer'),dfs)
    res = res.dropna(how='all') # drop uneeded columns out into future
    res = pandasutil.fillna_downbet(res) # use this as above ffills incorrectly at end of timeseries

    return res


def monthly_mean(df):
    """
    Given a price series, calculate the monthly mean and return as columns of means over years
            1  2  3 .. 12
    2000    x  x  x .. x
    2001    x  x  x .. x

    :param df:
    :return:
    """
    monthly_mean = df.groupby(pd.Grouper(freq='MS')).mean()
    month_pivot = monthly_mean.groupby([monthly_mean.index.month, monthly_mean.index.year]).sum().unstack()
    return month_pivot


if __name__ == '__main__':
    pass